import pytest
from pokie_container.constants import (
    SVC_CONTAINER,
    TREE_DEFAULT,
    TENANT_DEFAULT,
    NODE_DEFAULT,
)
from pokie_container.dto import TreeTypeRecord, NodeTypeRecord, NodeRecord
from pokie_container.service import ContainerService


@pytest.fixture
def svc_container(pokie_service_manager) -> ContainerService:
    return pokie_service_manager.get(SVC_CONTAINER)  # type: ContainerService


class TestPlatform:
    def test_tree_type(self, svc_container):
        records = svc_container.get_tree_type_list()
        assert len(records) == 1
        assert isinstance(records[0], TreeTypeRecord)

        # add a new tree type
        id_type = 100
        label = "my custom tree type"
        record = svc_container.add_tree_type(id_type, label)
        assert record is not None
        assert isinstance(record, TreeTypeRecord)

        records = svc_container.get_tree_type_list()
        assert len(records) == 2
        new_record = records[1]
        assert new_record.id == id_type
        assert new_record.label == label

    def test_node_type(self, svc_container):
        records = svc_container.get_node_type_list(TREE_DEFAULT)
        assert len(records) == 1
        assert isinstance(records[0], NodeTypeRecord) is True

        # add a custom node type
        id_type = 10
        label = "my custom node type"

        node_type = svc_container.add_node_type(TREE_DEFAULT, id_type, label)
        assert node_type is not None
        assert isinstance(node_type, NodeTypeRecord) is True

        records = svc_container.get_node_type_list(TREE_DEFAULT)
        assert len(records) == 2
        assert isinstance(records[1], NodeTypeRecord) is True

        node = records[1]
        assert node.id == id_type
        assert node.label == label

        result = svc_container.remove_node_type(id_type)
        assert result is True
        records = svc_container.get_node_type_list(TREE_DEFAULT)
        assert len(records) == 1

        # remove non-existing type
        result = svc_container.remove_node_type(id_type)
        assert result is False

    def test_node(self, svc_container: ContainerService):
        # custom node type
        id_node_type = 10
        node_type_label = "custom type"
        record = svc_container.add_node_type(
            TREE_DEFAULT, id_node_type, node_type_label
        )
        assert record is not None

        # create 10 nodes at root level, 5 nodes at second level
        nodes = {}
        for i in range(0, 10):
            record = svc_container.add_node(
                0, id_node_type, "root node {}".format(str(i))
            )
            nodes[record.id] = []
            for j in range(0, 2):
                child = svc_container.add_node(
                    0,
                    id_node_type,
                    "child node {} for node {}".format(str(j), str(i)),
                    id_parents=record.id,
                )
                nodes[record.id].append(child.id)

        for parent, children in nodes.items():
            record = svc_container.get_node(parent)
            node_record = svc_container.get_node_record(parent)
            assert isinstance(node_record, NodeRecord)
            assert node_record.id == parent

            # validate response
            for key in ["record", "parents", "children"]:
                assert key in record.keys()

            assert len(record["parents"]) == 0
            assert len(record["children"]) == 2
            for id_child in record["children"]:
                child = svc_container.get_node(id_child)
                # validate response
                for key in ["record", "parents", "children"]:
                    assert key in child.keys()

                assert len(child["parents"]) == 1
                assert child["parents"][0] == record["record"].id
                assert len(child["children"]) == 0

    def test_node_tree_multilevel(self, svc_container: ContainerService):
        # custom node type
        id_node_type = 10
        node_type_label = "custom type"
        record = svc_container.add_node_type(
            TREE_DEFAULT, id_node_type, node_type_label
        )
        assert record is not None

        # create a tree
        #       A
        #      / \
        #     B   C
        #    / \
        #   D   E
        node_a = svc_container.add_node(0, id_node_type, "node A")
        node_b = svc_container.add_node(0, id_node_type, "node B", id_parents=node_a.id)
        node_c = svc_container.add_node(0, id_node_type, "node C", id_parents=node_a.id)
        node_d = svc_container.add_node(0, id_node_type, "node D", id_parents=node_b.id)
        node_e = svc_container.add_node(0, id_node_type, "node E", id_parents=node_b.id)

        # validate node A
        read_a = svc_container.get_node(node_a.id)
        assert read_a["record"].label == "node A"
        assert len(read_a["parents"]) == 0
        assert len(read_a["children"]) == 2
        for node in [node_b, node_c]:
            assert node.id in read_a["children"]

        # validate node B
        read_b = svc_container.get_node(node_b.id)
        assert read_b["record"].label == "node B"
        assert len(read_b["parents"]) == 1
        assert read_b["parents"][0] == node_a.id
        assert len(read_b["children"]) == 2
        for node in [node_d, node_e]:
            assert node.id in read_b["children"]

        # validate node C
        read_c = svc_container.get_node(node_c.id)
        assert read_c["record"].label == "node C"
        assert len(read_c["parents"]) == 1
        assert read_c["parents"][0] == node_a.id
        assert len(read_c["children"]) == 0

        # validate node D
        read_d = svc_container.get_node(node_d.id)
        assert read_d["record"].label == "node D"
        assert len(read_d["parents"]) == 1
        assert read_d["parents"][0] == node_b.id
        assert len(read_d["children"]) == 0

        # validate node E
        read_e = svc_container.get_node(node_e.id)
        assert read_e["record"].label == "node E"
        assert len(read_e["parents"]) == 1
        assert read_e["parents"][0] == node_b.id
        assert len(read_e["children"]) == 0

        # extend the tree
        #     B
        #    / \
        #   D   E
        #    \ /
        #     F
        node_f = svc_container.add_node(
            0, id_node_type, "node F", id_parents=[node_d.id, node_e.id]
        )

        # validate node F
        read_f = svc_container.get_node(node_f.id)
        assert read_f["record"].label == "node F"
        assert len(read_f["parents"]) == 2
        for node in [node_d, node_e]:
            assert node.id in read_f["parents"]
        assert len(read_f["children"]) == 0

        # re-validate node E
        read_e = svc_container.get_node(node_e.id)
        assert read_e["record"].label == "node E"
        assert len(read_e["parents"]) == 1
        assert read_e["parents"][0] == node_b.id
        assert len(read_e["children"]) == 1
        assert read_e["children"][0] == node_f.id

    def test_node_tree_children_parent(self, svc_container: ContainerService):
        # create a tree
        #       A
        #      / \
        #     B   C
        #    / \
        #   D   E
        node_a = svc_container.add_node(0, NODE_DEFAULT, "node A")
        node_b = svc_container.add_node(0, NODE_DEFAULT, "node B", id_parents=node_a.id)
        node_c = svc_container.add_node(0, NODE_DEFAULT, "node C", id_parents=node_a.id)
        node_d = svc_container.add_node(0, NODE_DEFAULT, "node D", id_parents=node_b.id)
        node_e = svc_container.add_node(0, NODE_DEFAULT, "node E", id_parents=node_b.id)

        # check children
        children = svc_container.get_all_node_children_id(0, TREE_DEFAULT, node_b.id)
        assert len(children) == 2
        for i in [node_d.id, node_e.id]:
            assert i in children

        # check immediate children
        children = svc_container.get_children_id(0, TREE_DEFAULT, node_a.id)
        assert len(children) == 2
        for i in [node_b.id, node_c.id]:
            assert i in children

        # check parent
        parent = svc_container.get_all_node_parents_id(0, TREE_DEFAULT, node_e.id)
        assert len(parent) == 2
        for i in [node_a.id, node_b.id]:
            assert i in parent

        # check immediate parent
        parent = svc_container.get_parents_id(0, TREE_DEFAULT, node_e.id)
        assert len(parent) == 1
        for i in [node_b.id]:
            assert i in parent

        # check subtree
        parent = svc_container.get_node_subtree_id(0, TREE_DEFAULT, node_b.id)
        assert len(parent) == 3
        for i in [node_a.id, node_d.id, node_e.id]:
            assert i in parent

    def test_node_tree_multi_tree(self, svc_container: ContainerService):
        tree_1 = 10
        svc_container.add_tree_type(tree_1, "tree type 1")
        tree_2 = 20
        svc_container.add_tree_type(tree_2, "tree type 2")

        # custom node types
        tree1_id_node_type = 10
        tree1_node_type_label = "custom type for tree1"
        record = svc_container.add_node_type(
            tree_1, tree1_id_node_type, tree1_node_type_label
        )
        assert record is not None

        tree2_id_node_type = 11
        tree2_node_type_label = "custom type for tree2"
        record = svc_container.add_node_type(
            tree_2, tree2_id_node_type, tree2_node_type_label
        )
        assert record is not None

        # create first tree
        #       A
        #      / \
        #     B   C
        node_a1 = svc_container.add_node(0, tree1_id_node_type, "node A1")
        node_b1 = svc_container.add_node(
            0, tree1_id_node_type, "node B1", id_parents=node_a1.id
        )
        node_c1 = svc_container.add_node(
            0, tree1_id_node_type, "node C1", id_parents=node_a1.id
        )

        # create second tree
        #       A
        #      / \
        #     B   C
        node_a2 = svc_container.add_node(0, tree2_id_node_type, "node A2")
        node_b2 = svc_container.add_node(
            0, tree2_id_node_type, "node B2", id_parents=node_a2.id
        )
        node_c2 = svc_container.add_node(
            0, tree2_id_node_type, "node C2", id_parents=node_a2.id
        )

        # verify tree 1
        tree1 = svc_container.get_node_tree(0, tree_1)
        assert len(tree1) == 3
        for tid in [node_a1.id, node_b1.id, node_c1.id]:
            assert tid in tree1.keys()
        for tid in [node_a2.id, node_b2.id, node_c2.id]:
            assert tid not in tree1.keys()

        # verify tree 1 with records
        tree1 = svc_container.get_node_tree(0, tree_1, True)
        assert len(tree1) == 3
        tree1_ids = [k.id for k in tree1.keys()]
        for tid in [node_a1.id, node_b1.id, node_c1.id]:
            assert tid in tree1_ids
        for tid in [node_a2.id, node_b2.id, node_c2.id]:
            assert tid not in tree1_ids

        # verify tree 2
        tree2 = svc_container.get_node_tree(0, tree_2)
        assert len(tree2) == 3
        for tid in [node_a2.id, node_b2.id, node_c2.id]:
            assert tid in tree2.keys()
        for tid in [node_a1.id, node_b1.id, node_c1.id]:
            assert tid not in tree2.keys()

    def test_node_parent(self, svc_container: ContainerService):
        tree_1 = 10
        svc_container.add_tree_type(tree_1, "tree type 1")
        id_tree_2 = 20
        svc_container.add_tree_type(id_tree_2, "tree type 2")

        # custom node types
        tree1_id_node_type = 100
        tree1_node_type_label = "custom type for tree1"
        record = svc_container.add_node_type(
            tree_1, tree1_id_node_type, tree1_node_type_label
        )
        assert record is not None

        tree2_id_node_type = 110
        tree2_node_type_label = "custom type for tree2"
        record = svc_container.add_node_type(
            id_tree_2, tree2_id_node_type, tree2_node_type_label
        )
        assert record is not None

        # create first tree; used as fixture for multi-tree operation
        #       A  D
        #      / \
        #     B   C
        node_a1 = svc_container.add_node(0, tree1_id_node_type, "node A1")
        node_b1 = svc_container.add_node(
            0, tree1_id_node_type, "node B1", id_parents=node_a1.id
        )
        node_c1 = svc_container.add_node(
            0, tree1_id_node_type, "node C1", id_parents=node_a1.id
        )
        node_d1 = svc_container.add_node(0, tree1_id_node_type, "node D1")

        # create second tree
        #       A   D
        #      / \
        #     B   C
        node_a2 = svc_container.add_node(0, tree2_id_node_type, "node A2")
        node_b2 = svc_container.add_node(
            0, tree2_id_node_type, "node B2", id_parents=node_a2.id
        )
        node_c2 = svc_container.add_node(
            0, tree2_id_node_type, "node C2", id_parents=node_a2.id
        )
        node_d2 = svc_container.add_node(0, tree2_id_node_type, "node D2")

        # tree 2 parent manipulation
        # initial state: get node info
        node = svc_container.get_node(node_d2.id)
        assert isinstance(node, dict)
        assert isinstance(node["record"], NodeRecord)
        assert len(node["parents"]) == 0
        assert len(node["children"]) == 0

        # attempt to move node D to child of A
        # should fail, because they are on different levels
        result = svc_container.add_node_parent(node_d2.id, node_a2.id)
        assert result is False

        # attempt to remove node D
        result = svc_container.remove_node(node_d2.id)
        # and recreate as child of B
        node_d2 = svc_container.add_node(
            0, tree2_id_node_type, "node D2", id_parents=node_b2.id
        )
        node = svc_container.get_node(node_d2.id)
        assert isinstance(node, dict)
        assert isinstance(node["record"], NodeRecord)
        assert len(node["parents"]) == 1
        assert node_b2.id in node["parents"]
        assert len(node["children"]) == 0
        # add extra parent
        result = svc_container.add_node_parent(node_d2.id, node_c2.id)
        assert result is True
        node = svc_container.get_node(node_d2.id)
        assert len(node["parents"]) == 2
        assert node_b2.id in node["parents"]
        assert node_c2.id in node["parents"]
        # remove original parent
        result = svc_container.remove_node_parent(node_d2.id, node_b2.id)
        assert result is True
        node = svc_container.get_node(node_d2.id)
        assert len(node["parents"]) == 1
        assert node_c2.id in node["parents"]

        # get tree
        tree2 = svc_container.get_node_tree(0, id_tree_2)
        assert len(tree2) == 4
        assert len(tree2[node_d2.id]["children"]) == 0
        assert tree2[node_d2.id]["parents"] == [node_c2.id]

        # get tree nodes, no hierarchy
        tree2_nodes = svc_container.get_node_list(0, id_tree_2)
        assert len(tree2_nodes) == 4
        for r in tree2_nodes:
            assert r.id in [node_a2.id, node_b2.id, node_c2.id, node_d2.id]

    def test_node_src(self, svc_container: ContainerService):
        node_a = svc_container.add_node(0, NODE_DEFAULT, "node A1", src=100)
        node_b = svc_container.add_node(
            0, NODE_DEFAULT, "node B1", id_parents=node_a.id, src=101
        )
        node_c = svc_container.add_node(
            0, NODE_DEFAULT, "node C1", id_parents=node_a.id, src=102
        )
        node_d = svc_container.add_node(
            0, NODE_DEFAULT, "node D1", id_parents=node_b.id, src=103
        )

        node = svc_container.get_node_src(0, NODE_DEFAULT, 99)
        assert node is None
        node = svc_container.get_node_src(0, NODE_DEFAULT, 102)
        assert node is not None
        assert node.id == node_c.id
        node = svc_container.get_node_src(0, NODE_DEFAULT, 103)
        assert node is not None
        assert node.id == node_d.id



from rick_db import fieldmapper


@fieldmapper(tablename="node_tree_type", pk="id_node_tree_type")
class TreeTypeRecord:
    id = "id_node_tree_type"
    label = "label"


@fieldmapper(tablename="node_type", pk="id_node_type")
class NodeTypeRecord:
    id = "id_node_type"
    tree_type = "fk_node_tree_type"
    label = "label"


@fieldmapper(tablename="node", pk="id_node")
class NodeRecord:
    id = "id_node"
    tenant = "fk_tenant"
    tree_type = "fk_node_tree_type"
    node_type = "fk_node_type"
    created = "created_at"
    updated = "updated_at"
    src = "src"
    label = "label"
    attributes = "attributes"


@fieldmapper(tablename="node_tree", pk="id_node_tree")
class NodeTreeRecord:
    id = "id_node_tree"
    tenant = "fk_tenant"
    tree_type = "fk_node_tree_type"
    parent = "parent"
    child = "child"
    is_child = "is_child"
    depth = "depth"



from typing import List, Optional, Union

from pokie.constants import DI_DB
from rick.mixin import Injectable
from rick.util.datetime import iso8601_now

from pokie_container.dto import (
    TreeTypeRecord,
    NodeTypeRecord,
    NodeRecord,
    NodeTreeRecord,
)
from pokie_container.errors import NodeTypeNotFoundError, NodeNotFoundError, NodeError
from pokie_container.repository.repository import (
    NodeRepository,
    NodeTreeRepository,
    NodeTypeRepository,
    TreeTypeRepository,
)


class ContainerService(Injectable):
    def get_tree_type_list(self) -> List[TreeTypeRecord]:
        """
        Retrieve a list of tree types
        :return:
        """
        return self.repo_tree_type().get_all_ordered()

    def add_tree_type(self, id_tree_type: int, label: str) -> TreeTypeRecord:
        """
        Add a new tree type

        :param id_tree_type:
        :param label:
        :return:
        """
        record = TreeTypeRecord(
            id=id_tree_type,
            label=label,
        )
        self.repo_tree_type().insert(record)
        return record

    def get_node_type_list(self, id_tree_type: int) -> List[NodeTypeRecord]:
        """
        Retrieve list of node types
        :param id_tree_type:
        :return:
        """
        return self.repo_node_type().get_all_ordered(id_tree_type)

    def add_node_type(
        self, id_tree_type: int, id_node_type: int, label: str
    ) -> NodeTypeRecord:
        """
        Create a node type

        :param id_tree_type:
        :param id_node_type:
        :param label:
        :return:
        """
        record = NodeTypeRecord(id=id_node_type, tree_type=id_tree_type, label=label)
        self.repo_node_type().insert(record)
        return record

    def remove_node_type(self, id_node_type: int) -> bool:
        """
        Remove a node type

        :param id_node_type:
        :return:
        """
        record = self.repo_node_type().fetch_pk(id_node_type)
        if record is None:
            return False
        try:
            self.repo_node_type().delete_pk(id_node_type)
            return True
        except Exception as e:
            return False

    def get_node_record(self, id_node: int) -> Optional[NodeRecord]:
        """
        Get just the node record

        :param id_node:
        :return:
        """
        return self.repo_node().fetch_pk(id_node)

    def get_node(self, id_node: int) -> Optional[dict]:
        """
        Get node details

        Return format:
        {
            "record": NodeRecord,
            "parents": [list of immediate parent ids],
            "children": [list of immediate children ids],
        }

        :param id_node:
        :return:
        """
        record = self.repo_node().fetch_pk(id_node)  # type: NodeRecord
        if record is None:
            return None
        return {
            "record": record,
            "parents": self.repo_tree().get_parents(
                record.tenant, record.tree_type, record.id
            ),
            "children": self.repo_tree().get_children(
                record.tenant, record.tree_type, record.id
            ),
        }

    def get_node_src(
        self, id_tenant: int, id_tree_type: int, src: int
    ) -> Optional[NodeRecord]:
        """
        Find a node by src

        :param id_tenant:
        :param id_tree_type:
        :param src:
        :return:
        """
        return self.repo_node().find_by_src(id_tenant, id_tree_type, src)

    def get_all_node_children_id(
        self, id_tenant: int, id_tree_type: int, id_node: int
    ) -> List[int]:
        """
        Get all node ids for the children subtree of id_node

        :param id_tenant:
        :param id_tree_type:
        :param id_node:
        :return:
        """
        return self.repo_tree().get_all_children(id_tenant, id_tree_type, id_node)

    def get_all_node_parents_id(
        self, id_tenant: int, id_tree_type: int, id_node: int
    ) -> List[int]:
        """
        Get all node ids for the parents of the tree upto id_node

        :param id_tenant:
        :param id_tree_type:
        :param id_node:
        :return:
        """
        return self.repo_tree().get_all_parents(id_tenant, id_tree_type, id_node)

    def get_children_id(
        self, id_tenant: int, id_tree_type: int, id_node: int
    ) -> List[int]:
        """
        Get immediate children node ids

        :param id_tenant:
        :param id_tree_type:
        :param id_node:
        :return:
        """
        return self.repo_tree().get_children(id_tenant, id_tree_type, id_node)

    def get_parents_id(
        self, id_tenant: int, id_tree_type: int, id_node: int
    ) -> List[int]:
        """
        Get immediate parent node ids

        :param id_tenant:
        :param id_tree_type:
        :param id_node:
        :return:
        """
        return self.repo_tree().get_parents(id_tenant, id_tree_type, id_node)

    def get_node_subtree_id(
        self, id_tenant: int, id_tree_type: int, id_node: int
    ) -> List[int]:
        """
        Return all node ids in path above and below id_node

        :param id_tenant:
        :param id_tree_type:
        :param id_node:
        :return:
        """
        return self.repo_tree().get_all_children_parents(
            id_tenant, id_tree_type, id_node
        )

    def add_node(
        self,
        id_tenant: int,
        id_node_type: int,
        label: str,
        id_parents: Union[int, List] = None,
        src: int = None,
        attributes: dict = None,
    ) -> NodeRecord:
        """
        Creates a new node and adds it to the tree

        :param id_tenant:
        :param id_node_type:
        :param label:
        :param id_parents:
        :param src:
        :param attributes:
        :return:
        """
        node_type = self.repo_node_type().fetch_pk(id_node_type)  # type: NodeTypeRecord
        if not node_type:
            raise NodeTypeNotFoundError(
                "invalid node type: {}".format(str(id_node_type))
            )

        if not attributes:
            attributes = {}

        if id_parents is None:
            id_parents = []
        elif isinstance(id_parents, int):
            id_parents = [id_parents]

        now = iso8601_now()
        record = NodeRecord(
            tenant=id_tenant,
            tree_type=node_type.tree_type,
            node_type=node_type.id,
            created=now,
            updated=now,
            src=src,
            label=label,
            attributes=attributes,
        )
        record.id = self.repo_node().insert_pk(record)
        self.repo_tree().add_node(id_tenant, record.tree_type, record.id, id_parents)

        return record

    def add_node_parent(self, id_node: int, id_parent: int) -> bool:
        """
        Add an extra parent to a node
        Note: this cannot be used to move nodes around the tree, only to bind them to multiple parents

        :param id_node:
        :param id_parent:
        :return:
        """
        node = self.repo_node().fetch_pk(id_node)  # type: NodeRecord
        if node is None:
            return False
        return self.repo_tree().add_parent(
            node.tenant, node.tree_type, id_node, id_parent
        )

    def remove_node_parent(self, id_node: int, id_parent: int) -> bool:
        """
        Remove a parent from a node

        :param id_node:
        :param id_parent:
        :return:
        """
        node = self.repo_node().fetch_pk(id_node)  # type: NodeRecord
        if node is None:
            return False
        return self.repo_tree().delete_parent(
            node.tenant, node.tree_type, id_node, id_parent
        )

    def update_node(self, record: NodeRecord):
        """
        Updates a node record
        :param record:
        :return:
        """
        record.updated = iso8601_now()
        return self.repo_node().update(record)

    def remove_node(self, id_node: int):
        """
        Removes a node from the node table and the tree

        :param id_node:
        :return:
        """
        repo_node = self.repo_node()
        node = repo_node.fetch_pk(id_node)  # type: NodeRecord
        if not node:
            raise NodeNotFoundError("invalid node id {}".format(str(id_node)))

        if not self.repo_tree().delete_node(node.tenant, node.tree_type, id_node):
            raise NodeError("cannot remove node id {}".format(str(id_node)))

        self.repo_node().delete_pk(id_node)

    def get_node_tree(
        self, id_tenant: int, id_tree_type: int, fetch_records=False
    ) -> dict:
        """
        Fetches a tree

        the result has the following structure when fetch_records == False:
        {
          node_id: {
            'children': [list of node id's],
            'parents': [list of node id's],
            'nodes': [list of node id's],
            'depth': int
            }
          (...)
        }

        the result has the following structure when fetch_records == True:
        {
          node object: {
            'children': [list of node id's],
            'parents': [list of node id's],
            'nodes': [list of node id's],
            'depth': int
            }
          (...)
        }

        :param id_tenant:
        :param id_tree_type:
        :param fetch_records:
        :return: dict
        """
        result = self.repo_tree().get_tree(id_tenant, id_tree_type)
        if not fetch_records:
            return result

        record_result = {}
        repo = self.repo_node()
        for k, v in result.items():
            record_result[repo.fetch_pk(k)] = v
        return record_result

    def get_node_list(self, id_tenant: int, id_tree_type: int) -> List[NodeRecord]:
        """
        Retrieve a list of nodes sorted by label for a given tree
        :param id_tenant:
        :param id_tree_type:
        :return:
        """
        return self.repo_node().get_all_ordered(id_tenant, id_tree_type)

    def repo_node(self) -> NodeRepository:
        return NodeRepository(self.get_di().get(DI_DB))

    def repo_node_type(self) -> NodeTypeRepository:
        return NodeTypeRepository(self.get_di().get(DI_DB))

    def repo_tree_type(self) -> TreeTypeRepository:
        return TreeTypeRepository(self.get_di().get(DI_DB))

    def repo_tree(self) -> NodeTreeRepository:
        return NodeTreeRepository(self.get_di().get(DI_DB))



from typing import Union, List, Optional

from rick.util.datetime import iso8601_now
from rick_db import Repository
from rick_db.sql import Literal, Select

from pokie_container.dto import (
    TreeTypeRecord,
    NodeTypeRecord,
    NodeRecord,
    NodeTreeRecord,
)


class TreeTypeRepository(Repository):
    def __init__(self, db):
        super().__init__(db, TreeTypeRecord)

    def get_all_ordered(self) -> List[TreeTypeRecord]:
        qry = self.select().order(TreeTypeRecord.label)
        return self.fetch(qry)


class NodeTypeRepository(Repository):
    def __init__(self, db):
        super().__init__(db, NodeTypeRecord)

    def get_all_ordered(self, id_tree_type: int) -> List[NodeTypeRecord]:
        qry = (
            self.select()
            .where(NodeTypeRecord.tree_type, "=", id_tree_type)
            .order(NodeTypeRecord.label)
        )
        return self.fetch(qry)


class NodeRepository(Repository):
    def __init__(self, db):
        super().__init__(db, NodeRecord)

    def get_all_ordered(self, id_tenant: int, id_tree_type: int) -> List[NodeRecord]:
        qry = (
            self.select()
            .where(NodeRecord.tenant, "=", id_tenant)
            .where(NodeRecord.tree_type, "=", id_tree_type)
            .order(NodeRecord.label)
        )
        return self.fetch(qry)

    def find_by_src(
        self, id_tenant: int, id_tree_type: int, src: int
    ) -> Optional[NodeRecord]:
        qry = (
            self.select()
            .where(NodeRecord.tenant, "=", id_tenant)
            .where(NodeRecord.tree_type, "=", id_tree_type)
            .where(NodeRecord.src, "=", src)
        )
        return self.fetch_one(qry)


class NodeTreeRepository(Repository):
    def __init__(self, db):
        super().__init__(db, NodeTreeRecord)

    def add_node(
        self,
        id_tenant: int,
        id_tree_type: int,
        id_node: int,
        id_parent_node: Union[List, int] = 0,
    ) -> bool:
        if isinstance(id_parent_node, int):
            id_parent_node = [id_parent_node]

        depth = 0
        self._db.begin()
        for id_parent in id_parent_node:
            qry = (
                self.select()
                .where(NodeTreeRecord.tenant, "=", id_tenant)
                .where(NodeTreeRecord.tree_type, "=", id_tree_type)
                .where(NodeTreeRecord.child, "=", id_parent)
            )
            for row in self.fetch(qry):  # type: NodeTreeRecord
                if row.depth >= depth:
                    depth = row.depth + 1
                record = NodeTreeRecord(
                    tenant=id_tenant,
                    tree_type=id_tree_type,
                    parent=row.parent,
                    child=id_node,
                    is_child=True if (row.parent == row.child) else False,
                    depth=row.depth,
                )
                self.insert(record)

        record = NodeTreeRecord(
            tenant=id_tenant,
            tree_type=id_tree_type,
            parent=id_node,
            child=id_node,
            is_child=False,
            depth=depth,
        )
        self.insert(record)
        self._db.commit()
        return True

    def get_by_node(
        self, id_tenant: int, id_tree_type: int, id_node: int
    ) -> Optional[NodeTreeRecord]:
        qry = (
            self.select()
            .where(NodeTreeRecord.tenant, "=", id_tenant)
            .where(NodeTreeRecord.tree_type, "=", id_tree_type)
            .where(NodeTreeRecord.child, "=", id_node)
            .where(NodeTreeRecord.parent, "=", id_node)
        )
        return self.fetch_one(qry)

    def get_parents(self, id_tenant: int, id_tree_type: int, id_node: int) -> List:
        """
        Get immediate parent node ids
        :param id_tenant:
        :param id_tree_type:
        :param id_node:
        :return:
        """
        qry = (
            self.select(cols=[NodeTreeRecord.parent])
            .where(NodeTreeRecord.tenant, "=", id_tenant)
            .where(NodeTreeRecord.tree_type, "=", id_tree_type)
            .where(NodeTreeRecord.child, "=", id_node)
            .where(NodeTreeRecord.is_child, "=", True)
        )
        result = []
        for row in self.fetch(qry):
            result.append(row.parent)
        return result

    def get_children(self, id_tenant: int, id_tree_type: int, id_node: int) -> List:
        """
        Get immediate children node ids
        :param id_tenant:
        :param id_tree_type:
        :param id_node:
        :return:
        """
        qry = (
            self.select(cols=[NodeTreeRecord.child])
            .where(NodeTreeRecord.tenant, "=", id_tenant)
            .where(NodeTreeRecord.tree_type, "=", id_tree_type)
            .where(NodeTreeRecord.parent, "=", id_node)
            .where(NodeTreeRecord.is_child, "=", True)
        )
        result = []
        for row in self.fetch(qry):
            result.append(row.child)
        return result

    def add_parent(
        self, id_tenant: int, id_tree_type: int, id_node: int, id_parent: int
    ) -> bool:
        """
        Add a parent to a node
        Note: this cannot be used to move nodes between levels; parent must be at depth-1 from the current node

        :param id_tenant:
        :param id_tree_type:
        :param id_node:
        :param id_parent:
        :return:
        """
        if id_parent in self.get_parents(id_tenant, id_tree_type, id_node):
            return False

        node = self.get_by_node(id_tenant, id_tree_type, id_node)
        parent = self.get_by_node(id_tenant, id_tree_type, id_parent)
        if node is None or parent is None:
            return False

        if node.depth != parent.depth + 1:
            return False

        self._db.begin()
        subqry = (
            self.select(cols=NodeTreeRecord.parent)
            .where(NodeTreeRecord.tenant, "=", id_tenant)
            .where(NodeTreeRecord.tree_type, "=", id_tree_type)
            .where(NodeTreeRecord.child, "=", id_node)
        )

        qry = (
            self.select()
            .where(NodeTreeRecord.tenant, "=", id_tenant)
            .where(NodeTreeRecord.tree_type, "=", id_tree_type)
            .where(NodeTreeRecord.child, "=", id_parent)
            .where(NodeTreeRecord.parent, "not in", subqry)
        )
        for row in self.fetch(qry):
            record = NodeTreeRecord(
                tenant=id_tenant,
                tree_type=id_tree_type,
                parent=row.parent,
                child=id_node,
                is_child=True if (row.parent == row.child) else False,
                depth=row.depth,
            )
            self.insert(record)
        self._db.commit()
        return True

    def delete_parent(
        self, id_tenant: int, id_tree_type: int, id_node: int, id_parent: int
    ) -> bool:
        """
        Remove a parent from a node

        :param id_tenant:
        :param id_tree_type:
        :param id_node:
        :param id_parent:
        :return:
        """
        parents = self.get_parents(id_tenant, id_tree_type, id_node)
        if id_parent not in parents:
            return False
        parents.remove(id_parent)

        # cannot remove last parent
        if len(parents) == 0:
            return False

        toks = [str(p) for p in parents]
        parents = ",".join(toks)

        subqry = (
            self.select(cols=NodeTreeRecord.parent)
            .where(NodeTreeRecord.tenant, "=", id_tenant)
            .where(NodeTreeRecord.tree_type, "=", id_tree_type)
            .where(NodeTreeRecord.child, "in", Literal("({})".format(parents)))
        )

        qry = (
            self.select()
            .where(NodeTreeRecord.tenant, "=", id_tenant)
            .where(NodeTreeRecord.tree_type, "=", id_tree_type)
            .where(NodeTreeRecord.child, "=", id_node)
            .where(NodeTreeRecord.parent, "<>", id_node)
            .where(NodeTreeRecord.parent, "not in ", subqry)
        )

        self._db.begin()
        for row in self.fetch(qry):
            self.delete_where(
                [
                    (NodeTreeRecord.tenant, "=", id_tenant),
                    (NodeTreeRecord.tree_type, "=", id_tree_type),
                    (NodeTreeRecord.parent, "=", row.parent),
                    (NodeTreeRecord.child, "=", row.child),
                    (NodeTreeRecord.depth, "=", row.depth),
                ]
            )
        self._db.commit()
        return True

    def get_all_children(self, id_tenant: int, id_tree_type: int, id_node: int) -> List:
        """
        Get all node ids for the children subtree of id_node
        :param id_tenant:
        :param id_tree_type:
        :param id_node:
        :return:
        """
        qry = (
            self.select(cols=[NodeTreeRecord.child])
            .where(NodeTreeRecord.tenant, "=", id_tenant)
            .where(NodeTreeRecord.tree_type, "=", id_tree_type)
            .where(NodeTreeRecord.parent, "=", id_node)
            .where(NodeTreeRecord.child, "<>", id_node)
        )
        result = []
        for row in self.fetch(qry):
            result.append(row.child)
        return result

    def get_all_parents(self, id_tenant: int, id_tree_type: int, id_node: int) -> List:
        """
        Get all node ids for the parents of the tree upto id_node

        :param id_tenant:
        :param id_tree_type:
        :param id_node:
        :return:
        """
        qry = (
            self.select(cols=[NodeTreeRecord.parent])
            .where(NodeTreeRecord.tenant, "=", id_tenant)
            .where(NodeTreeRecord.tree_type, "=", id_tree_type)
            .where(NodeTreeRecord.child, "=", id_node)
            .where(NodeTreeRecord.parent, "<>", id_node)
        )
        result = []
        for row in self.fetch(qry):
            result.append(row.parent)
        return result

    def get_all_children_parents(
        self, id_tenant: int, id_tree_type: int, id_node: int
    ) -> List:
        """
        Return all node ids in path above and below id_node

        :param id_tenant:
        :param id_tree_type:
        :param id_node:
        :return:
        """
        result = self.get_children(id_tenant, id_tree_type, id_node)
        result.extend(self.get_parents(id_tenant, id_tree_type, id_node))
        return result

    def delete_node(self, id_tenant: int, id_tree_type: int, id_node: int) -> bool:
        """
        Removes a node from the tree
        The node must not have children

        :param id_tenant:
        :param id_tree_type:
        :param id_node:
        :return:
        """
        children = self.get_children(id_tenant, id_tree_type, id_node)
        if len(children) > 0:
            return False

        self.delete_where(
            [
                (NodeTreeRecord.tenant, "=", id_tenant),
                (NodeTreeRecord.tree_type, "=", id_tree_type),
                (NodeTreeRecord.child, "=", id_node),
            ]
        )
        return True

    def get_tree(self, id_tenant: int, id_tree_type: int):
        """
        Get complete tree in iterable format

        the result has the following structure:
        {
          node_id: {
            'children': [list of node id's],
            'parents': [list of node id's],
            'nodes': [list of node id's],
            'depth': int
            }
          (...)
        }

        :param id_tenant:
        :param id_tree_type:
        :return:
        """
        tn = self._tablename
        qry = (
            Select(self._dialect)
            .from_(
                {tn: "a"},
                cols={
                    NodeTreeRecord.parent: None,
                    NodeTreeRecord.depth: None,
                    Literal(
                        "array(SELECT child FROM {tn} as sub WHERE sub.parent = a.parent AND is_child )".format(
                            tn=tn
                        )
                    ): "children",
                    Literal(
                        "array(SELECT parent FROM {tn} as sub WHERE sub.child = a.parent AND is_child )".format(
                            tn=tn
                        )
                    ): "parents",
                    Literal(
                        "array(SELECT child FROM {tn} as sub WHERE sub.parent = a.parent )".format(
                            tn=tn
                        )
                    ): "nodes",
                },
                schema=self._schema,
            )
            .where(NodeTreeRecord.tenant, "=", id_tenant)
            .where(NodeTreeRecord.tree_type, "=", id_tree_type)
            .where(NodeTreeRecord.parent, "=", Literal("a.child"))
            .order(NodeTreeRecord.depth)
        )
        result = {}
        with self._db.cursor() as c:
            sql, values = qry.assemble()
            for row in c.fetchall(sql, values):
                result[row["parent"]] = {
                    "children": row["children"],
                    "parents": row["parents"],
                    "nodes": row["nodes"],
                    "depth": row["depth"],
                }
        return result



# Version
POKIE_CONTAINER_VERSION = ["0", "9", "0"]


def get_version():
    return ".".join(POKIE_CONTAINER_VERSION)


SVC_CONTAINER = "pokie_container.container"

TENANT_DEFAULT = 0  # default tenant
TREE_DEFAULT = 1  # default tree type
NODE_DEFAULT = 1  # default node type
